@@ -396,7 +396,7 @@ public final class Generator {
 
             if (!state.allowNaN()) {
                 if (state.strict() && state.getAsJSON() != null) {
-                    IRubyObject castedValue = state.getAsJSON().call(context, object);
+                    IRubyObject castedValue = state.getAsJSON().call(context, object, context.getRuntime().getFalse());
                     if (castedValue != object) {
                         getHandlerFor(context.runtime, castedValue).generate(context, session, castedValue, buffer);
                         return;
@@ -606,6 +606,8 @@ public final class Generator {
     }
 
     private static void processEntry(ThreadContext context, Session session, OutputStream buffer, RubyHash.RubyHashEntry entry, boolean firstPair, ByteList objectNl, byte[] indent, ByteList spaceBefore, ByteList space) {
+        StringEncoder encoder = session.getStringEncoder(context);
+
         IRubyObject key = (IRubyObject) entry.getKey();
         IRubyObject value = (IRubyObject) entry.getValue();
 
@@ -619,11 +621,11 @@ public final class Generator {
             Ruby runtime = context.runtime;
 
             IRubyObject keyStr = castKey(context, key);
-            if (keyStr == null || !(keyStr instanceof RubyString)) {
+            if (keyStr == null || !(keyStr instanceof RubyString) || !encoder.hasValidEncoding((RubyString)keyStr)) {
                 GeneratorState state = session.getState(context);
                 if (state.strict()) {
                     if (state.getAsJSON() != null) {
-                        key = state.getAsJSON().call(context, key);
+                        key = state.getAsJSON().call(context, key, context.getRuntime().getTrue());
                         keyStr = castKey(context, key);
                     }
 
@@ -664,6 +666,19 @@ public final class Generator {
 
         @Override
         void generate(ThreadContext context, Session session, RubyString object, OutputStream buffer) throws IOException {
+            GeneratorState state = session.getState(context);
+            StringEncoder encoder = session.getStringEncoder(context);
+
+            if (state.strict() && !encoder.hasValidEncoding(object) && state.getAsJSON() != null) {
+                IRubyObject value = state.getAsJSON().call(context, object, context.getRuntime().getFalse());
+                if (value instanceof RubyString) {
+                    object = (RubyString)value;
+                } else {
+                    Handler handler = getHandlerFor(context.runtime, value);
+                    handler.generate(context, session, value, buffer);
+                    return;
+                }
+            }
             generateString(context, session, object, buffer);
         }
     }
@@ -760,7 +775,7 @@ public final class Generator {
         GeneratorState state = session.getState(context);
         if (state.strict()) {
             if (state.getAsJSON() != null) {
-                IRubyObject value = state.getAsJSON().call(context, object);
+                IRubyObject value = state.getAsJSON().call(context, object, context.getRuntime().getFalse());
                 Handler handler = getHandlerFor(context.runtime, value);
                 if (handler == GENERIC_HANDLER) {
                     throw Utils.buildGeneratorError(context, object, value + " returned by as_json not allowed in JSON").toThrowable();
