@@ -27,9 +27,10 @@ public class SWARBasicStringEncoder extends StringEncoder {
         // slice a string, the underlying byte array is the same, but the
         // begin index and real size are different. When reading from the ptrBytes
         // array, we need to always add ptr to the index.
-        ByteBuffer bb = ByteBuffer.wrap(ptrBytes, ptr, len);
+        ByteBuffer bb = ByteBuffer.wrap(ptrBytes, 0, ptr + len);
+
         while (pos + 8 <= len) {
-            long x = bb.getLong(pos);
+            long x = bb.getLong(ptr + pos);
             if (skipChunk(x)) {
                 pos += 8;
                 continue;
@@ -48,7 +49,7 @@ public class SWARBasicStringEncoder extends StringEncoder {
         }
 
         if (pos + 4 <= len) {
-            int x = bb.getInt(pos);
+            int x = bb.getInt(ptr + pos);
             if (skipChunk(x)) {
                 pos += 4;
             }
@@ -70,7 +71,7 @@ public class SWARBasicStringEncoder extends StringEncoder {
         }
     }
 
-    private boolean skipChunk(long x) {
+    boolean skipChunk(long x) {
         long is_ascii = 0x8080808080808080L & ~x;
         long xor2 = x ^ 0x0202020202020202L;
         long lt32_or_eq34 = xor2 - 0x2121212121212121L;
@@ -79,7 +80,7 @@ public class SWARBasicStringEncoder extends StringEncoder {
         return ((lt32_or_eq34 | eq92) & is_ascii) == 0;
     }
 
-    private boolean skipChunk(int x) {
+    boolean skipChunk(int x) {
         int is_ascii = 0x80808080 & ~x;
         int xor2 = x ^ 0x02020202;
         int lt32_or_eq34 = xor2 - 0x21212121;
