@@ -47,6 +47,19 @@ module JSON
 
       SCRIPT_SAFE_ESCAPE_PATTERN = /[\/"\\\x0-\x1f\u2028-\u2029]/
 
+      def self.native_type?(value) # :nodoc:
+        (false == value || true == value || nil == value || String === value || Array === value || Hash === value || Integer === value || Float === value || Fragment === value)
+      end
+
+      def self.native_key?(key) # :nodoc:
+        (Symbol === key || String === key)
+      end
+
+      def self.valid_encoding?(string) # :nodoc:
+        return false unless string.encoding == ::Encoding::UTF_8 || string.encoding == ::Encoding::US_ASCII
+        string.is_a?(Symbol) || string.valid_encoding?
+      end
+
       # Convert a UTF8 encoded Ruby string _string_ to a JSON string, encoded with
       # UTF16 big endian characters as \u????, and return it.
       def self.utf8_to_json(string, script_safe = false) # :nodoc:
@@ -204,7 +217,7 @@ module JSON
           return if @max_nesting.zero?
           current_nesting = depth + 1
           current_nesting > @max_nesting and
-            raise NestingError, "nesting of #{current_nesting} is too deep"
+            raise NestingError, "nesting of #{current_nesting} is too deep. Did you try to serialize objects with circular references?"
         end
 
         # Returns true, if circular data structures are checked,
@@ -339,6 +352,10 @@ module JSON
           dup.generate(obj, anIO)
         end
 
+        private def initialize_copy(_orig)
+          @depth = 0
+        end
+
         # Handles @allow_nan, @buffer_initial_length, other ivars must be the default value (see above)
         private def generate_json(obj, buf)
           case obj
@@ -422,6 +439,8 @@ module JSON
 
         # Return the value returned by method +name+.
         def [](name)
+          ::JSON.deprecation_warning("JSON::State#[] is deprecated and will be removed in json 3.0.0")
+
           if respond_to?(name)
             __send__(name)
           else
@@ -431,6 +450,8 @@ module JSON
         end
 
         def []=(name, value)
+          ::JSON.deprecation_warning("JSON::State#[]= is deprecated and will be removed in json 3.0.0")
+
           if respond_to?(name_writer = "#{name}=")
             __send__ name_writer, value
           else
@@ -448,10 +469,10 @@ module JSON
             state = State.from_state(state) if state
             if state&.strict?
               value = self
-              if state.strict? && !(false == value || true == value || nil == value || String === value || Array === value || Hash === value || Integer === value || Float === value || Fragment === value)
+              if state.strict? && !Generator.native_type?(value)
                 if state.as_json
-                  value = state.as_json.call(value)
-                  unless false == value || true == value || nil == value || String === value || Array === value || Hash === value || Integer === value || Float === value || Fragment === value
+                  value = state.as_json.call(value, false)
+                  unless Generator.native_type?(value)
                     raise GeneratorError.new("#{value.class} returned by #{state.as_json} not allowed in JSON", value)
                   end
                   value.to_json(state)
@@ -509,13 +530,17 @@ module JSON
               end
               result << state.indent * depth if indent
 
-              if state.strict? && !(Symbol === key || String === key)
-                if state.as_json
-                  key = state.as_json.call(key)
+              if state.strict?
+                if state.as_json && (!Generator.native_key?(key) || !Generator.valid_encoding?(key))
+                  key = state.as_json.call(key, true)
                 end
 
-                unless Symbol === key || String === key
-                  raise GeneratorError.new("#{key.class} not allowed as object key in JSON", value)
+                unless Generator.native_key?(key)
+                  raise GeneratorError.new("#{key.class} not allowed as object key in JSON", key)
+                end
+
+                unless Generator.valid_encoding?(key)
+                  raise GeneratorError.new("source sequence is illegal/malformed utf-8", key)
                 end
               end
 
@@ -527,10 +552,10 @@ module JSON
               end
 
               result = +"#{result}#{key_json}#{state.space_before}:#{state.space}"
-              if state.strict? && !(false == value || true == value || nil == value || String === value || Array === value || Hash === value || Integer === value || Float === value || Fragment === value)
+              if state.strict? && !Generator.native_type?(value)
                 if state.as_json
-                  value = state.as_json.call(value)
-                  unless false == value || true == value || nil == value || String === value || Array === value || Hash === value || Integer === value || Float === value || Fragment === value
+                  value = state.as_json.call(value, false)
+                  unless Generator.native_type?(value)
                     raise GeneratorError.new("#{value.class} returned by #{state.as_json} not allowed in JSON", value)
                   end
                   result << value.to_json(state)
@@ -588,10 +613,10 @@ module JSON
             each { |value|
               result << delim unless first
               result << state.indent * depth if indent
-              if state.strict? && !(false == value || true == value || nil == value || String === value || Array === value || Hash === value || Integer === value || Float === value || Fragment === value || Symbol == value)
+              if state.strict? && !Generator.native_type?(value)
                 if state.as_json
-                  value = state.as_json.call(value)
-                  unless false == value || true == value || nil == value || String === value || Array === value || Hash === value || Integer === value || Float === value || Fragment === value || Symbol === value
+                  value = state.as_json.call(value, false)
+                  unless Generator.native_type?(value)
                     raise GeneratorError.new("#{value.class} returned by #{state.as_json} not allowed in JSON", value)
                   end
                   result << value.to_json(state)
@@ -625,7 +650,7 @@ module JSON
               if state.allow_nan?
                 to_s
               elsif state.strict? && state.as_json
-                casted_value = state.as_json.call(self)
+                casted_value = state.as_json.call(self, false)
 
                 if casted_value.equal?(self)
                   raise GeneratorError.new("#{self} not allowed in JSON", self)
@@ -662,14 +687,25 @@ module JSON
           # \u????.
           def to_json(state = nil, *args)
             state = State.from_state(state)
-            if encoding == ::Encoding::UTF_8
-              unless valid_encoding?
+            string = self
+
+            if state.strict? && state.as_json
+              unless Generator.valid_encoding?(string)
+                string = state.as_json.call(string, false)
+                unless string.is_a?(::String)
+                  return string.to_json(state, *args)
+                end
+              end
+            end
+
+            if string.encoding == ::Encoding::UTF_8
+              unless string.valid_encoding?
                 raise GeneratorError.new("source sequence is illegal/malformed utf-8", self)
               end
-              string = self
             else
-              string = encode(::Encoding::UTF_8)
+              string = string.encode(::Encoding::UTF_8)
             end
+
             if state.ascii_only?
               %("#{JSON::TruffleRuby::Generator.utf8_to_json_ascii(string, state.script_safe)}")
             else
