@@ -183,7 +183,9 @@ class JSONGeneratorTest < Test::Unit::TestCase
     assert_equal('{"1":2}', json)
     s = JSON.state.new
     assert s.check_circular?
-    assert s[:check_circular?]
+    assert_deprecated_warning(/JSON::State/) do
+      assert s[:check_circular?]
+    end
     h = { 1=>2 }
     h[3] = h
     assert_raise(JSON::NestingError) {  generate(h) }
@@ -193,7 +195,9 @@ class JSONGeneratorTest < Test::Unit::TestCase
     a << a
     assert_raise(JSON::NestingError) {  generate(a, s) }
     assert s.check_circular?
-    assert s[:check_circular?]
+    assert_deprecated_warning(/JSON::State/) do
+      assert s[:check_circular?]
+    end
   end
 
   def test_falsy_state
@@ -375,28 +379,32 @@ class JSONGeneratorTest < Test::Unit::TestCase
   end
 
   def test_hash_likeness_set_symbol
-    state = JSON.state.new
-    assert_equal nil, state[:foo]
-    assert_equal nil.class, state[:foo].class
-    assert_equal nil, state['foo']
-    state[:foo] = :bar
-    assert_equal :bar, state[:foo]
-    assert_equal :bar, state['foo']
-    state_hash = state.to_hash
-    assert_kind_of Hash, state_hash
-    assert_equal :bar, state_hash[:foo]
+    assert_deprecated_warning(/JSON::State/) do
+      state = JSON.state.new
+      assert_equal nil, state[:foo]
+      assert_equal nil.class, state[:foo].class
+      assert_equal nil, state['foo']
+      state[:foo] = :bar
+      assert_equal :bar, state[:foo]
+      assert_equal :bar, state['foo']
+      state_hash = state.to_hash
+      assert_kind_of Hash, state_hash
+      assert_equal :bar, state_hash[:foo]
+    end
   end
 
   def test_hash_likeness_set_string
-    state = JSON.state.new
-    assert_equal nil, state[:foo]
-    assert_equal nil, state['foo']
-    state['foo'] = :bar
-    assert_equal :bar, state[:foo]
-    assert_equal :bar, state['foo']
-    state_hash = state.to_hash
-    assert_kind_of Hash, state_hash
-    assert_equal :bar, state_hash[:foo]
+    assert_deprecated_warning(/JSON::State/) do
+      state = JSON.state.new
+      assert_equal nil, state[:foo]
+      assert_equal nil, state['foo']
+      state['foo'] = :bar
+      assert_equal :bar, state[:foo]
+      assert_equal :bar, state['foo']
+      state_hash = state.to_hash
+      assert_kind_of Hash, state_hash
+      assert_equal :bar, state_hash[:foo]
+    end
   end
 
   def test_json_state_to_h_roundtrip
@@ -822,29 +830,44 @@ class JSONGeneratorTest < Test::Unit::TestCase
 
   def test_json_generate_as_json_convert_to_proc
     object = Object.new
-    assert_equal object.object_id.to_json, JSON.generate(object, strict: true, as_json: :object_id)
+    assert_equal object.object_id.to_json, JSON.generate(object, strict: true, as_json: -> (o, is_key) { o.object_id })
+  end
+
+  def assert_float_roundtrip(expected, actual)
+    assert_equal(expected, JSON.generate(actual))
+    assert_equal(actual, JSON.parse(JSON.generate(actual)), "JSON: #{JSON.generate(actual)}")
   end
 
   def test_json_generate_float
-      values = [-1.0, 1.0, 0.0, 12.2, 7.5 / 3.2, 12.0, 100.0, 1000.0]
-      expecteds = ["-1.0", "1.0", "0.0", "12.2", "2.34375", "12.0", "100.0", "1000.0"]
+    assert_float_roundtrip "-1.0", -1.0
+    assert_float_roundtrip "1.0", 1.0
+    assert_float_roundtrip "0.0", 0.0
+    assert_float_roundtrip "12.2", 12.2
+    assert_float_roundtrip "2.34375", 7.5 / 3.2
+    assert_float_roundtrip "12.0", 12.0
+    assert_float_roundtrip "100.0", 100.0
+    assert_float_roundtrip "1000.0", 1000.0
 
-      if RUBY_ENGINE == "jruby"
-        values << 1746861937.7842371
-        expecteds << "1.7468619377842371E9"
-      else
-        values << 1746861937.7842371
-        expecteds << "1746861937.7842371"
-      end
+    if RUBY_ENGINE == "jruby"
+      assert_float_roundtrip "1.7468619377842371E9", 1746861937.7842371
+    else
+      assert_float_roundtrip "1746861937.7842371", 1746861937.7842371
+    end
 
-      if RUBY_ENGINE == "ruby"
-        values << -2.2471348024634545e-08 << -2.2471348024634545e-09 << -2.2471348024634545e-10
-        expecteds << "-0.000000022471348024634545" << "-0.0000000022471348024634545" << "-2.2471348024634546e-10"
-      end
+    if RUBY_ENGINE == "ruby"
+      assert_float_roundtrip "100000000000000.0", 100000000000000.0
+      assert_float_roundtrip "1e+15", 1e+15
+      assert_float_roundtrip "-100000000000000.0", -100000000000000.0
+      assert_float_roundtrip "-1e+15", -1e+15
+      assert_float_roundtrip "1111111111111111.1", 1111111111111111.1
+      assert_float_roundtrip "1.1111111111111112e+16", 11111111111111111.1
+      assert_float_roundtrip "-1111111111111111.1", -1111111111111111.1
+      assert_float_roundtrip "-1.1111111111111112e+16", -11111111111111111.1
 
-      values.zip(expecteds).each do |value, expected|
-        assert_equal expected, value.to_json
-      end
+      assert_float_roundtrip "-0.000000022471348024634545", -2.2471348024634545e-08
+      assert_float_roundtrip "-0.0000000022471348024634545", -2.2471348024634545e-09
+      assert_float_roundtrip "-2.2471348024634546e-10", -2.2471348024634545e-10
+    end
   end
 
   def test_numbers_of_various_sizes
